divert(-1)
#********************************************************************
#  File PROTOKEN.M4
#  M4 definitions for implementing the Retargetable Concurrent Small C
#  virtual machine on the Parallax Propeller
#
#  Remember: c file name must be DOS 8.3 format !!!
#
#  compile c program: rcsc2 program.c
#  this will create program.m4
#  compile m4 to spin: m4 protoken.m4 program.m4 > program.spin
#
#December 27, 2007 Peter Verkaik
#
#********************************************************************
#
#RCSC pseudo instructions
#
#********************************************************************
define(`CSCEXTRN',`{{Retargetable Concurrent Small C compiler}}


CON

  STACKSIZE = 100

  {{ comparison attributes }}
  EQ   =    1   {{ 0x01 equal }}
  GE   =  156   {{ 0x9C greater or equal }}
  GEu  =   28   {{ 0x1C greater or equal (unsigned) }}
  GT   =  140   {{ 0x8C greater than }}
  GTu  =   12   {{ 0x0C greater than (unsigned) }}
  LE   =  154   {{ 0x9A less or equal }}
  LEu  =   26   {{ 0x1A less or equal (unsigned) }}
  LT   =  138   {{ 0x8A less than }}
  LTu  =   10   {{ 0x0A less than (unsigned) }}
  NE   =   40   {{ 0x28 not equal }}

  EQf  =    1   {{ 0x01 equal }}
  GEf  =  156   {{ 0x9C greater or equal }}
  GTf  =  140   {{ 0x8C greater than }}
  LEf  =  154   {{ 0x9A less or equal }}
  LTf  =  138   {{ 0x8A less than }}
  NEf  =   40   {{ 0x28 not equal }}

  {{ pcode followed by 0 operands }}
  PCODE_OPER0 = 0
  pNOP     =   0   {{ no operation }}
  pADD12   =   1   {{ add sr to pr }}
  pADD21   =   2   {{ add pr to sr }}
  pAND12   =   3   {{ AND sr to pr }}
  pANEG1   =   4   {{ arith negate pr }}
  pASL12   =   5   {{ arith shift left sr by pr into pr }}
  pASR12   =   6   {{ arith shift right sr by pr into pr }}
  pCALL1   =   7   {{ call function thru pr }}
  pCLI     =   8   {{ interrupt disable }}
  pCOM1    =   9   {{ ones complement pr }}
  pCSCINIT =  10   {{ concurrent Small C Initialization }}
  pDBL1    =  11   {{ double pr }}
  pDBL2    =  12   {{ double sr }}
  pDIV12   =  13   {{ div pr by sr }}
  pDIV12u  =  14   {{ div pr by sr unsigned }}
  pENTER   =  15   {{ set stack frame on function entry }}
  pLNEG1   =  16   {{ logical negate pr }}
  pMOD12   =  17   {{ modulo pr by sr }}
  pMOD12u  =  18   {{ modulo pr by sr unsigned }}
  pMOVE21  =  19   {{ move pr to sr }}
  pMUL12   =  20   {{ multiply pr by sr }}
  pMUL12u  =  21   {{ multiply pr by sr unsigned }}
  pOR12    =  22   {{ OR sr onto pr }}
  pPOP2    =  23   {{ pop stack into sr }}
  pPUSH1   =  24   {{ push pr onto stack }}
  pPUSH2   =  25   {{ push sr }}
  pPUTbp1  =  26   {{ put pr byte in mem thru sr ptr }}
  pPUTwp1  =  27   {{ put pr word in mem thru sr ptr }}
  pRETURN  =  28   {{ restore stack and return }}
  pSTI     =  29   {{ interrupt enable }}
  pSUB12   =  30   {{ sub sr from pr }}
  pSWAP12  =  31   {{ swap pr and sr }}
  pSWAP1s  =  32   {{ swap pr and top of stack }}
  pSWITCH  =  33   {{ find switch case }}
  pXOR12   =  34   {{ XOR pr with sr }}

  {{ pcode followed by 1 operand }}
  PCODE_OPER1 = 35
  pADDbpn  =  35   {{ add n to mem byte thru sr ptr }}
  pADDSP   =  36   {{ add to stack pointer }}
  pARGCNTn =  37   {{ pass arg count to function }}
  pCMP12   =  38   {{ set pr to (sr CMP pr) }}
  pGETb1p  =  39   {{ get byte into pr from mem thru sr ptr }}
  pGETb1pu =  40   {{ get unsigned byte into pr from mem thru sr ptr }}
  pGETb1s  =  41   {{ get byte into pr from stack }}
  pGETb1su =  42   {{ get unsigned byte into pr from stack }}
  pGETw1p  =  43   {{ get word into pr from mem thru sr ptr }}
  pGETw1s  =  44   {{ get word into pr from stack }}
  pGETw2p  =  45   {{ get word into sr thru sr ptr }}
  pGETw2s  =  46   {{ get word into sr from stack }}
  pPOINT1s =  47   {{ point pr to stack item }}
  pPOINT2s =  48   {{ point sr to stack }}
  pPUSHp   =  49   {{ push word from mem thru sr ptr }}
  pPUSHs   =  50   {{ push word from stack }}

  {{ pcode followed by 2 operands }}
  PCODE_OPER2 = 51
  pADD1n   =  51   {{ add n to pr }}
  pADD2n   =  52   {{ add immediate to sr }}
  pADDwpn  =  53   {{ add n to mem word thru sr ptr }}
  pANDbm1  =  54   {{ bitwise and pr with mem byte thru label }}
  pANDwm1  =  55   {{ bitwise and pr with mem word thru label }}
  pCALLm   =  56   {{ call function directly }}
  pGETb1m  =  57   {{ get byte into pr from mem thru label }}
  pGETb1mu =  58   {{ get unsigned byte into pr from mem thru label }}
  pGETw1m  =  59   {{ get word into pr from mem thru label }}
  pGETw1n  =  60   {{ get word of value n into pr }}
  pGETw2m  =  61   {{ get word into sr from mem (label) }}
  pGETw2n  =  62   {{ get word of value n into sr }}
  pJMPm    =  63   {{ jump to label }}
  pORbm1   =  64   {{ bitwise or pr with mem byte thru label }}
  pORwm1   =  65   {{ bitwise or pr with mem word thru label }}
  pPOINT1m =  66   {{ point pr to mem item thru label }}
  pPOINT2m =  67   {{ point sr to mem thru label }}
  pPUSHm   =  68   {{ push word from mem thru label }}
  pPUTbm1  =  69   {{ put pr byte in mem thru label }}
  pPUTwm1  =  70   {{ put pr word in mem thru label }}
  pXORbm1  =  71   {{ bitwise xor pr with mem byte thru label }}
  pXORwm1  =  72   {{ bitwise xor pr with mem word thru label }}

  {{ pcode followed by 3 operands }}
  PCODE_OPER3 = 73
  pADDbmn  =  73   {{ add n to mem byte thru label }}
  pANDbmn  =  74   {{ bitwise and n with mem byte thru label }}
  pCMP10f  =  75   {{ jump to label if (pr CMP 0) is false }}
  pORbmn   =  76   {{ bitwise or n with mem byte thru label }}
  pPUTbmn  =  77   {{ put n into mem byte thru label }}
  pXORbmn  =  78   {{ bitwise xor n with mem byte thru label }}

  {{ pcode followed by 4 operands }}
  PCODE_OPER4 = 79
  pADDwmn  =  79   {{ add n to mem word thru label }}
  pANDwmn  =  80   {{ bitwise and n with mem word thru label }}
  pORwmn   =  81   {{ bitwise or n with mem word thru label }}
  pPUTwmn  =  82   {{ put n into mem word thru label }}
  pXORwmn  =  83   {{ bitwise xor n with mem word thru label }}
')

define(`CSCTERM',
`PUB start

  VM_SP := @VM_STACK + STACKSIZE  {{initialize stack }}
  VM_PC := @_main                 {{initialize program counter }}

  {{ VM mainloop }}
  `repeat'  
    vm_pcode := byte[VM_PC++]              {{ pcode }}
    `if' vm_pcode => PCODE_OPER1
      vm_DX.byte[0] := byte[VM_PC++]       {{ operand 1 }}
      `if' vm_pcode => PCODE_OPER2
        vm_DX.byte[1] := byte[VM_PC++]     {{ operand 2 }}
        `if' vm_pcode => PCODE_OPER3
          vm_CX.byte[0] := byte[VM_PC++]   {{ operand 3 }}
          `if' vm_pcode => PCODE_OPER4
            vm_CX.byte[1] := byte[VM_PC++] {{ operand 4 }}
            {{pcode operators with 4 operands}}
            case VM_PCODE
              pADDwmn:  {{ add n to memory word thru label }}
              pANDwmn:  {{ bitwise and n with mem word thru label }}
              pORwmn:   {{ bitwise or n with mem word thru label }}
              pPUTwmn:  {{ put n into memory word thru label }}
              pXORwmn:  {{ bitwise xor n with mem word thru label }}
          else
            {{pcode operators with 3 operands}}
            case VM_PCODE
              pADDbmn:  {{ add n to memory byte thru label }}
              pANDbmn:  {{ bitwise and n with memory byte }}
              pCMP10f:  {{ jump to label if (pr CMP 0) is false }}
              pORbmn:   {{ bitwise or n with memory byte }}
              pPUTbmn:  {{ put n into memory byte thru label }}
              pXORbmn:  {{ bitwise xor n with memory byte }}
        else
          {{pcode operators with 2 operands}}
          case VM_PCODE
            pADD1n:   {{ add n to pr }}
            pADD2n:   {{ add n to sr }}
            pADDwpn:  {{ add n to memory word thru sr pointer }}
            pANDbm1:  {{ bitwise and pr with mem byte thru label }}
            pANDwm1:  {{ bitwise and pr with mem word thru label }}
            pCALLm:   {{ call function directly }}
            pGETb1m:  {{ get byte into pr from memory thru label }}
            pGETb1mu: {{ get unsigned byte into pr from memory thru label }}
            pGETw1m:  {{ get word into pr from memory thru label }}
            pGETw1n:  {{ get word of value n into pr }}
            pGETw2m:  {{ get word into sr from memory thru label }}
            pGETw2n:  {{ get word of value n into sr }}
            pJMPm:    {{ jump to label }}
            pORbm1:   {{ bitwise or pr with mem byte thru label }}
            pORwm1:   {{ bitwise or pr with mem word thru label }}
            pPOINT1m: {{ point pr to memory item thru label }}
            pPOINT2m: {{ point sr to memory item thru label }}
            pPUSHm:   {{ push word from memory thru label }}
            pPUTbm1:  {{ put pr byte into memory thru label }}
            pPUTwm1:  {{ put pr word into memory thru label }}
            pXORbm1:  {{ bitwise xor pr with mem byte thru label }}
            pXORwm1:  {{ bitwise xor pr with mem word thru label }}
      else
        {{pcode operators with 1 operand}}
        case VM_PCODE
          pADDbpn:  {{ add n to memory byte through sr pointer }}
          pADDSP:   {{ add n to stackpointer }}
          pARGcntn: {{ pass argument count to function }}
          pCMP12:   {{ set pr to (sr CMP pr) }}
          pGETb1p:  {{ get byte into pr from memory thru sr pointer }}
          pGETb1pu: {{ get unsigned byte into pr from memory thru sr pointer }}
          pGETb1s:  {{ get byte into pr from stack }}
          pGETb1su: {{ get unsigned byte into pr from stack }}
          pGETw1p:  {{ get word into pr from memory thru sr pointer }}
          pGETw1s:  {{ get word into pr from stack }}
          pGETw2p:  {{ get word into sr from memory thru sr pointer }}
          pGETw2s:  {{ get word into sr from stack }}
          pPOINT1s: {{ point pr to stack item }}
          pPOINT2s: {{ point sr to stack item }}
          pPUSHp:   {{ push word from memory thru sr pointer }}
          pPUSHs:   {{ push word from stack }}
    else
      {{pcode operators with 0 operands}}
      case VM_PCODE
        pNOP:     {{ no operation }}
        pADD12:   {{ add sr to pr }}
        pADD21:   {{ add pr to sr }}
        pAND12:   {{ and pr with sr }}
        pANEG1:   {{ arithmetically negate pr }}
        pASL12:   {{ arithmetically shift left sr by pr into pr }}
        pASR12:   {{ arithmetically shift right sr by pr into pr }}
        pCALL1:   {{ call function thru pr }}
        pCLI:     {{ disable interrupt }}
        pCOM1:    {{ one complement pr }}
        pCSCINIT: {{ initialize csc }}
        pDBL1:    {{ double pr }}
        pDBL2:    {{ double sr }}
        pDIV12:   {{ divide pr by sr }}
        pDIV12u:  {{ divide pr by sr (unsigned) }}
        pENTER:   {{ set stack frame upon function entry }}
        pLNEG1:   {{ logical negation of pr }}
        pMOD12:   {{ modulo pr by sr }}
        pMOD12u:  {{ modulo pr by sr (unsigned) }}
        pMOVE21:  {{ move pr to sr }}
        pMUL12:   {{ multiply pr by sr }}
        pMUL12u:  {{ multiply pr by sr (unsigned) }}
        pOR12:    {{ or pr with sr }}
        pPOP2:    {{ pop stack into sr }}
        pPUSH1:   {{ push pr onto stack }}
        pPUSH2:   {{ push sr onto stack }}
        pPUTBP1:  {{ put pr byte into memory thru sr pointer }}
        pPUTWP1:  {{ put pr word into memory thru sr pointer }}
        pRETURN:  {{ restore stack and return }}
        pSTI:     {{ enable interrupt }}
        pSUB12:   {{ subtract sr from pr }}
        pSWAP12:  {{ swap pr and sr }}
        pSWAP1s:  {{ swap pr and top of stack }}
        pSWITCH:  {{ scan switch table to find case }}
        pXOR12:   {{ xor pr with sr }}

`'')

define(`EXTERNAL',
`{{`include' $1()}}')

define(`TOSEG',
`define(`CURSEG', $1) ifelse($1, 1, `',`')')

define(`ENDSEG',
`ifelse($1, 1,`',`')')

define(`DECLPUBLIC',
`ifelse(CURSEG, 2, `
$1', `
DECLARE($1)')')

define(`DECLARE', 
`define(`LABELON', 1)
DECLAREMEM($1,')dnl

define(`REFm', 
`define(`LABELON', 1)
REFmMEM($1,')dnl

define(`SHADOW', `')
define(`MODULENAME', `')
define(`BEGINLIT', `')
define(`ENDLIT', `')
define(`BEGINDUMP', `')dnl

#********************************************************************
#
#macros used internally
#
#********************************************************************
define(`repeat', 
`ifelse(eval($2 <= 0), 1, , 
`repeat(`$1', decr($2))'
$1)')

define(`DECLAREMEM', 
`define(`LABELON', 0)
ifelse($2, _WORD_, `$1		word 0', 
$2, _WORDn, `$1		word 0',
$2, _WORDr0, `$1		word 0[$3]',
$2, _BYTE_, `$1		byte 0', 
$2, _BYTEn, `$1		byte 0',
`$1		byte 0[$3]')')

define(`REFmMEM', 
`define(`LABELON', 0)
ifelse($2, _WORD_, `$1		byte	', 
$2, _WORDn, `$1		byte	$3',
$2, _WORDr0, `$1		byte	$3',
$2, _BYTE_, `$1		byte	', 
$2, _BYTEn, `$1		byte	$3',
`$1		byte	$3')')

#********************************************************************
#
#interrupt, task and monitor table creation macros
#
#********************************************************************
define(`INTRS', 
`	byte	$1,$2')

define(`TASKS', 
`	byte	$1,$2,$3')

define(`MNTRS',
`	byte	$1')

#********************************************************************
#
#RCSC virtual machine instruction set
#
#********************************************************************
#
# Generates tokenized output.
# PCODE constant is followed by max. 3 operands.
# The last operand may be a 12bits word. (Others are 8bit).

define(`ADD12',
`	byte	pADD12')

define(`ADD1n',
`	byte	pADD1n,($1)&255,($1)>>8')

define(`ADD21',
`	byte	pADD21')

define(`ADD2n',
`	byte	pADD2n,($1)&255,($1)>>8')

define(`ADDbpn',
`	byte	pADDbpn,($1)&255')

define(`ADDwpn',
`	byte	pADDwpn,($1)&255,($1)>>8')

define(`ADDm_',
`ADDm_COMMAn($1')

define(`ADDm_COMMAn',
`	byte	pADDwmn,(@$1)&255,(@$1)>>8,($2)&255,($2)>>8')

define(`ADDSP',
`	byte	pADDSP,($1)&255')

define(`AND12',
`	byte	pAND12')

define(`ANEG1',
`	byte	pANEG1')

define(`ARGCNTn',
`	byte	pARGCNTn,($1)&255')

define(`ASL12',
`	byte	pASL12')

define(`ASR12',
`	byte	pASR12')

define(`CALL1',
`	byte	pCALL1')

define(`CALLm',
`	byte	pCALLm,(@$1)&255,(@$1)>>8')

define(`BYTE_', 
`ifelse(LABELON, 1, `_BYTE_, $1)', `	byte	')')

define(`BYTEn', 
`ifelse(LABELON, 1, `_BYTEn, $1)', `	byte	$1')')

define(`BYTEr0', 
`ifelse(LABELON, 1, `_BYTEr0, $1)', `	byte	$1,0')')

define(`COM1',
`	byte	pCOM1')

define(`COMMAn',
`,$1)')

define(`DBL1',
`	byte	pDBL1')

define(`DBL2',
`	byte	pDBL2')

define(`DECbp',
`	byte	pADDbpn,(-1)&255')

define(`DECwp',
`	byte	pADDwpn,(-1)&255,(-1)>>8')

define(`DIV12',
`	byte	pDIV12')

define(`DIV12u',
`	byte	pDIV12u')

define(`ENTER',
`	byte	pENTER')

define(`EQ10f',
`	byte	pCMP10f,(@$1)&255,(@$1)>>8,EQf')

define(`EQ12',
`	byte	pCMP12,EQ')

define(`GE10f',
`	byte	pCMP10f,(@$1)&255,(@$1)>>8,GEf')

define(`GE12',
`	byte	pCMP12,GE')

define(`GE12u',
`	byte	pCMP12,GEu')

define(`GETb1m',
`	byte	pGETb1m,(@$1)&255,(@$1)>>8')

define(`GETb1mu',
`	byte	pGETb1mu,(@$1)&255,(@$1)>>8')

define(`GETb1p',
`	byte	pGETb1p,($1)&255')

define(`GETb1pu',
`	byte	pGETb1pu,($1)&255')

define(`GETb1s',
`	byte	pGETb1s,($1)&255')

define(`GETb1su',
`	byte	pGETb1su,($1)&255')

define(`GETw1m',
`	byte	pGETw1m,(@$1)&255,(@$1)>>8')

define(`GETw1m_',
`GETw1m_PLUSn($1')

define(`GETw1m_PLUSn',
`	byte	pGETw1m,(@$1)&255,(@$1)>>8')

define(`GETw1n',
`	byte	pGETw1n,($1)&255,($1)>>8')

define(`GETw1p',
`	byte	pGETw1p,($1)&255')

define(`GETw1s',
`	byte	pGETw1s,($1)&255')

define(`GETw2m',
`	byte	pGETw2m,(@$1)&255,(@$1)>>8')

define(`GETw2n',
`	byte	pGETw2n,($1)&255,($1)>>8')

define(`GETw2p',
`	byte	pGETw2p,($1)&255')

define(`GETw2s',
`	byte	pGETw2s,($1)&255')

define(`GT10f',
`	byte	pCMP10f,(@$1)&255,(@$1)>>8,GTf')

define(`GT12',
`	byte	pCMP12,GT')

define(`GT12u',
`	byte	pCMP12,GTu')

define(`INCbp',
`	byte	pADDbpn,(1)&255')

define(`INCwp',
`	byte	pADDwpn,(1)&255,(1)>>8')

define(`WORD_', 
`ifelse(LABELON, 1, `_WORD_)', `	byte	')')

define(`WORDn', 
`ifelse(LABELON, 1, `_WORDn, $1)', `	byte	($1)&255,($1)>>8')')

define(`WORDr0', 
`ifelse(LABELON, 1, `_WORDr0, $1)', `	dfw	$1')')

define(`JMPm',
`	byte	pJMPm,(@$1)&255,(@$1)>>8')

define(`LABm',
`$1')

define(`LE10f',
`	byte	pCMP10f,(@$1)&255,(@$1)>>8,LEf')

define(`LE12',
`	byte	pCMP12,LE')

define(`LE12u',
`	byte	pCMP12,LEu')

define(`LNEG1',
`	byte	pLNEG1')

define(`LT10f',
`	byte	pCMP10f,(@$1)&255,(@$1)>>8,LTf')

define(`LT12',
`	byte	pCMP12,LT')

define(`LT12u',
`	byte	pCMP12,LTu')

define(`MOD12',
`	byte	pMOD12')

define(`MOD12u',
`	byte	pMOD12u')

define(`MOVE21',
`	byte	pMOVE21')

define(`MUL12',
`	byte	pMUL12')

define(`MUL12u',
`	byte	pMUL12u')

define(`NE10f',
`	byte	pCMP10f,(@$1)&255,(@$1)>>8,NEf')

define(`NE12',
`	byte	pCMP12,NE')

define(`NEARm',
`	byte	(@$1)&255,(@$1)>>8')

define(`OR12',
`	byte	pOR12')

define(`PLUSn',
`+$1')

define(`POINT1l',
`	byte	pPOINT1m,(@$1)&255,(@$1)>>8')

define(`POINT1m',
`	byte	pPOINT1m,(@$1)&255,(@$1)>>8')

define(`POINT1s',
`	byte	pPOINT1s,($1)&255')

define(`POINT2m',
`	byte	pPOINT2m,(@$1)&255,(@$1)>>8')

define(`POINT2m_',
`POINT2m_PLUSn($1')

define(`POINT2m_PLUSn',
`	byte	pPOINT2m,(@$1)&255,(@$1)>>8')

define(`POINT2s',
`	byte	pPOINT2s,($1)&255')

define(`POP2',
`	byte	pPOP2')

define(`PUSH1',
`	byte	pPUSH1')

define(`PUSH2',
`	byte	pPUSH2')

define(`PUSHm',
`	byte	pPUSHm,(@$1)&255,(@$1)>>8')

define(`PUSHp',
`	byte	pPUSHp,($1)&255')

define(`PUSHs',
`	byte	pPUSHs,($1)&255')

define(`PUT_m_',
`PUT_m_COMMAn($1')dnl

define(`PUT_m_COMMAn',
`	byte	pPUTwmn,(@$1)&255,(@$1)>>8,($2)&255,($2)>>8')

define(`PUTbm1',
`	byte	pPUTbm1,(@$1)&255,(@$1)>>8')

define(`PUTbp1',
`	byte	pPUTbp1')

define(`PUTwm1',
`	byte	pPUTwm1,(@$1)&255,(@$1)>>8')

define(`PUTwp1',
`	byte	pPUTwp1')

define(`rDEC1',
`	byte	pADD1n,(-($1))&255,(-($1))>>8')

define(`rDEC2',
`	byte	pADD2n,(-($1))&255,(-($1))>>8')

define(`RETURN',
`	byte	pRETURN')

define(`rINC1',
`	byte	pADD1n,($1)&255,($1)>>8')

define(`rINC2',
`	byte	pADD2n,($1)&255,($1)>>8')

define(`SUB_m_',
`SUB_m_COMMAn($1')

define(`SUB_m_COMMAn',
`	byte	pADDwmn,(@$1)&255,(@$1)>>8,(-($2))&255,(-($2))>>8')

define(`SUB12',
`	byte	pSUB12')

define(`SUB1n',
`	byte	pADD1n,(-($1))&255,(-($1))>>8')

define(`SUBbpn',
`	byte	pADDbpn,(-($1))&255')

define(`SUBwpn',
`	byte	pADDwpn,(-($1))&255,(-($1))>>8')

define(`SWAP12',
`	byte	pSWAP12')

define(`SWAP1s',
`	byte	pSWAP1s')

define(`SWITCH',
`	byte	pSWITCH')

define(`XOR12',
`	byte	pXOR12')

define(`CSCINIT',
`	byte	pCSCINIT')

define(`ENTERM',
`	byte	pCLI')

define(`EXITM',
`	byte	pSTI')

divert
